@@ -31,7 +31,7 @@ from numpy.testing import (assert_equal, assert_array_equal,
 from numpy import shape, allclose, array_equal, ravel, isnan, isinf
 
 import numexpr
-from numexpr import E, NumExpr, evaluate, re_evaluate, disassemble, use_vml
+from numexpr import E, NumExpr, evaluate, re_evaluate, validate, disassemble, use_vml
 from numexpr.expressions import ConstantNode
 
 import unittest
@@ -370,13 +370,43 @@ class test_evaluate(TestCase):
         assert_array_equal(x, array([86., 124., 168.]))
 
     def test_re_evaluate_dict(self):
+        a1 = array([1., 2., 3.])
+        b1 = array([4., 5., 6.])
+        c1 = array([7., 8., 9.])
+        local_dict={'a': a1, 'b': b1, 'c': c1}
+        x = evaluate("2*a + 3*b*c", local_dict=local_dict)
+        x = re_evaluate(local_dict=local_dict)
+        assert_array_equal(x, array([86., 124., 168.]))
+
+    def test_validate(self):
         a = array([1., 2., 3.])
         b = array([4., 5., 6.])
         c = array([7., 8., 9.])
-        x = evaluate("2*a + 3*b*c", local_dict={'a': a, 'b': b, 'c': c})
+        retval = validate("2*a + 3*b*c")
+        assert(retval is None)
         x = re_evaluate()
         assert_array_equal(x, array([86., 124., 168.]))
 
+    def test_validate_missing_var(self):
+        a = array([1., 2., 3.])
+        b = array([4., 5., 6.])
+        retval = validate("2*a + 3*b*c")
+        assert(isinstance(retval, KeyError))
+
+    def test_validate_syntax(self):
+        retval = validate("2+")
+        assert(isinstance(retval, SyntaxError))
+
+    def test_validate_dict(self):
+        a1 = array([1., 2., 3.])
+        b1 = array([4., 5., 6.])
+        c1 = array([7., 8., 9.])
+        local_dict={'a': a1, 'b': b1, 'c': c1}
+        retval = validate("2*a + 3*b*c", local_dict=local_dict)
+        assert(retval is None)
+        x = re_evaluate(local_dict=local_dict)
+        assert_array_equal(x, array([86., 124., 168.]))
+
     # Test for issue #22
     def test_true_div(self):
         x = arange(10, dtype='i4')
@@ -474,11 +504,63 @@ class test_evaluate(TestCase):
         a = arange(3)
         try:
             evaluate("a < [0, 0, 0]")
-        except TypeError:
+        except (ValueError, TypeError):
             pass
         else:
             self.fail()
 
+    def test_forbidden_tokens(self):
+        # Forbid dunder
+        try:
+            evaluate('__builtins__')
+        except ValueError:
+            pass
+        else:
+            self.fail()
+
+        # Forbid colon for lambda funcs
+        try: 
+            evaluate('lambda x: x')
+        except ValueError:
+            pass
+        else:
+            self.fail()
+
+        # Forbid indexing
+        try:
+            evaluate('locals()[]')
+        except ValueError:
+            pass
+        else:
+            self.fail()
+
+        # Forbid semicolon
+        try:
+            evaluate('import os;')
+        except ValueError:
+            pass
+        else:
+            self.fail()
+
+        # Attribute access
+        try:
+            evaluate('os.cpucount()')
+        except ValueError:
+            pass
+        else:
+            self.fail()
+
+        # But decimal point must pass
+        a = 3.0
+        evaluate('a*2.')
+        evaluate('2.+a')
+        
+
+        
+
+
+
+
     def test_disassemble(self):
         assert_equal(disassemble(NumExpr(
             "where(m, a, -1)", [('m', bool), ('a', float)])),
@@ -511,6 +593,30 @@ class test_evaluate(TestCase):
         evaluate('3', out=a)
         assert_equal(a, 3)
 
+    def test_negative_mod(self):
+        # Test for issue #413, modulus of negative integers. C modulus is 
+        # actually remainder op, and hence different from Python modulus.
+        a = np.array([-500, -135, 0, 0, 135, 500], dtype=np.int32)
+        n = np.array([-360, -360, -360, 360, 360, 360], dtype=np.int32)
+        out_i = evaluate('a % n')
+        assert_equal(out_i, np.mod(a, n))
+
+        b = a.astype(np.int64)
+        m = n.astype(np.int64)
+        out_l = evaluate('b % m')
+        assert_equal(out_l, np.mod(b, m))
+
+    def test_negative_power_scalar(self):
+        # Test for issue #428, where the power is negative and the base is an
+        # integer. This was running afoul in the precomputation in `expressions.py:pow_op()`
+        base = np.array([-2, -1, 0, 1, 2, 3], dtype=np.int32)
+        out_i = evaluate('base ** -1.0')
+        assert_equal(out_i, np.power(base, -1.0))
+
+        base = np.array([-2, -1, 0, 1, 2, 3], dtype=np.int64)
+        out_l = evaluate('base ** -1.0')
+        assert_equal(out_l, np.power(base, -1.0))
+
 
     def test_ex_uses_vml(self):
         vml_funcs = [ "sin", "cos", "tan", "arcsin", "arccos", "arctan",
@@ -624,7 +730,7 @@ def equal(a, b, exact):
         return (shape(a) == shape(b)) and alltrue(ravel(a) == ravel(b), axis=0)
     else:
         if hasattr(a, 'dtype') and a.dtype == 'f4':
-            atol = 1e-5  # Relax precission for special opcodes, like fmod
+            atol = 1e-5  # Relax precision for special opcodes, like fmod
         else:
             atol = 1e-8
         return (shape(a) == shape(b) and
